﻿using System;
namespace KernelPerceptron
{
  class KernelPerceptronProgram
  {
    static void Main(string[] args)
    {
      Console.WriteLine("\nBegin kernel perceptron demo \n");
      int numFeatures = 2;

      Console.WriteLine("Goal is classification/prediction (-1/+1) with two predictors \n");
      Console.WriteLine("Setting up 21 training data items \n");

      // donut-like data (not linearly separable)
      double[][] trainData = new double[21][];
      trainData[0] = new double[] { 2.0, 3.0, -1 };
      trainData[1] = new double[] { 2.0, 5.0, -1 };
      trainData[2] = new double[] { 2.0, 7.0, -1 };
      trainData[3] = new double[] { 3.0, 2.0, -1 };
      trainData[4] = new double[] { 3.0, 8.0, -1 };
      trainData[5] = new double[] { 4.0, 2.0, -1 };
      trainData[6] = new double[] { 4.0, 8.0, -1 };
      trainData[7] = new double[] { 5.0, 2.0, -1 };
      trainData[8] = new double[] { 5.0, 8.0, -1 };
      trainData[9] = new double[] { 6.0, 3.0, -1 };
      trainData[10] = new double[] { 6.0, 5.0, -1 };
      trainData[11] = new double[] { 6.0, 7.0, -1 };
      trainData[12] = new double[] { 3.0, 4.0, +1 };
      trainData[13] = new double[] { 3.0, 5.0, +1 };
      trainData[14] = new double[] { 3.0, 6.0, +1 };
      trainData[15] = new double[] { 4.0, 4.0, +1 };
      trainData[16] = new double[] { 4.0, 5.0, +1 };
      trainData[17] = new double[] { 4.0, 6.0, +1 };
      trainData[18] = new double[] { 5.0, 4.0, +1 };
      trainData[19] = new double[] { 5.0, 5.0, +1 };
      trainData[20] = new double[] { 5.0, 6.0, +1 };
      int numTrain = trainData.Length;

      // test
      double[][] testData = new double[4][];
      testData[0] = new double[] { 2.0, 4.0, -1 };
      testData[1] = new double[] { 6.0, 6.0, -1 };
      testData[2] = new double[] { 3.5, 4.5, +1 };
      testData[3] = new double[] { 5.5, 5.5, +1 };
      //testData[3] = new double[] { 5, 6, +1 };

      Console.WriteLine("First training data is trainData[0] = (2.0, 3.0, -1) ");
      Console.WriteLine("Last training data is trainData[20] = (5.0, 6.0, +1) \n");
      double[] x = new double[numFeatures];  // inputs
      int[] a = new int[numTrain];  // 'wrong counters' one cell for each train data
      
      double sigma = 1.5;  // for the kernel function
      //double sigma = 0.5;  // gives 100% accuracy on train and test

      // train
      double[][] kernelMatrix = new double[numTrain][];  // pre-compute all Kernel
      for (int i = 0; i < kernelMatrix.Length; ++i)
        kernelMatrix[i] = new double[numTrain];
      for (int i = 0; i < numTrain; ++i)
      {
        for (int j = 0; j < numTrain; ++j)
        {
          double k = Kernel(trainData[i], trainData[j], sigma);
          kernelMatrix[i][j] = kernelMatrix[j][i] = k;
        }
      }

      Console.WriteLine("Starting kernel perceptron training, sigma = 1.5 ");
      int maxEpoch = 10;
      int epoch = 0;
      while (epoch < maxEpoch)
      {
        for (int i = 0; i < numTrain; ++i)  // each training data
        {
          int di = (int)trainData[i][numFeatures];  // get desired/correct class -1 or +1

          double y = 0.0; // weighted sum of kernel results
          for (int j = 0; j < numTrain; ++j) // each other train items
          {
            int dj = (int)trainData[j][numFeatures];  // 
            double kern = kernelMatrix[i][j];
            y += a[j] * dj * kern;
          } 

          // update curr wrong-counters if predicted y is wrong
          if ((di == -1 && y >= 0.0) || (di == 1 && y <= 0.0))
            ++a[i]; // increment wrong counter for this train data
        }  // i, each train data

        ++epoch;
      }  // while
      // end training

      Console.WriteLine("Training complete \n");

      Console.WriteLine("First two wrong-counters: ");
      Console.WriteLine("a[0] = " + a[0] + "   a[1] = " + a[1]);
      Console.WriteLine("Last two wrong-counters: ");
      Console.WriteLine("a[19] = " + a[19] + "  a[20] = " + a[20]);

      double trainAcc = Accuracy(trainData, a, trainData, sigma, false);  // silent
      Console.WriteLine("\nAccuracy on training data = " + trainAcc.ToString("F4"));

      Console.WriteLine("\nAnalyzing test data: ");
      double testAcc = Accuracy(testData, a, trainData, sigma, true);  // verbose
      //Console.WriteLine("Accuracy on test data     = " + testAcc.ToString("F4"));

      Console.WriteLine("\nEnd kernel perceptron demo \n");
      Console.ReadLine();

    } // Main

    static double Kernel(double[] d1, double[] d2, double sigma)
    {
      // RBF kernel. assumes class label in last cell
      double num = 0.0;
      for (int i = 0; i < d1.Length - 1; ++i)  // note -1
        num += (d1[i] - d2[i]) * (d1[i] - d2[i]);
      double denom = 2.0 * sigma * sigma;
      double z = num / denom;
      double result = Math.Exp(-z);
      return result;
    }

    static double Accuracy(double[][] data, int[] a, double[][] trainData, double sigma, bool verbose)
    {
      // assumes data is like 2.0, 3.5  +1 (actual class in last cell)
      int numFeatures = data[0].Length - 1;
      int numCorrect = 0;
      int numWrong = 0;

      for (int i = 0; i < data.Length; ++i)  // i is data to evaluate
      {
        int d = (int)data[i][numFeatures];  // get desired/correct class -1 or +1

        double y = 0.0; 
        for (int j = 0; j < trainData.Length; ++j) // use all train items
        {
          int d_other = (int)trainData[j][numFeatures];  // +1 or -1
          y += a[j] * d_other * Kernel(data[i], trainData[j], sigma);
        }
 
        if ((d == -1 && y >= 0.0) || (d == 1 && y <= 0.0))
          ++numWrong;
        else
          ++numCorrect;

        if (verbose == true)
        {
          Console.Write("Input: ");
          for (int j = 0; j < data[i].Length - 1; ++j)
            Console.Write(data[i][j].ToString("F1") + " ");
          if (data[i][data[i].Length - 1] == 1)
            Console.Write("+");
          Console.Write(data[i][data[i].Length - 1].ToString("F0"));

          Console.Write("   y calc = ");
          if (y >= 0.0) Console.Write("+"); 
          Console.Write(y.ToString("F4"));
          Console.Write("  predicted = ");
          if (y < 0.0) Console.Write("-1");
          else Console.Write("+1");

          if ((d == -1 && y >= 0.0) || (d == 1 && y <= 0.0))
            Console.WriteLine("  wrong");
          else
            Console.WriteLine("  correct");
        }

      }  // for each data
      return (1.0 * numCorrect) / (numCorrect + numWrong);
    }

  } // Program

} // ns
